@@ -6,7 +6,7 @@
  *	License as published by the Free Software Foundation version 2.1
  *	of the License.
  *
- * Copyright (c) 2003-2012 Thomas Graf <tgraf@suug.ch>
+ * Copyright (c) 2003-2008 Thomas Graf <tgraf@suug.ch>
  */
 
 /**
@@ -157,29 +157,29 @@ struct nl_object *nl_cache_get_prev(struct nl_object *obj)
 /** @} */
 
 /**
- * @name Cache Creation/Deletion
+ * @name Cache Allocation/Deletion
  * @{
  */
 
 /**
- * Allocate an empty cache
- * @arg ops		cache operations to base the cache on
- * 
- * @return A newly allocated and initialized cache.
+ * Allocate new cache
+ * @arg ops		Cache operations
+ *
+ * Allocate and initialize a new cache based on the cache operations
+ * provided.
+ *
+ * @return Allocated cache or NULL if allocation failed.
  */
 struct nl_cache *nl_cache_alloc(struct nl_cache_ops *ops)
 {
 	struct nl_cache *cache;
 
 	cache = calloc(1, sizeof(*cache));
-	if (!cache) {
-		nl_errno(ENOMEM);
+	if (!cache)
 		return NULL;
-	}
 
 	nl_init_list_head(&cache->c_items);
 	cache->c_ops = ops;
-	cache->c_refcnt = 1;
 
 	NL_DBG(2, "Allocated cache %p <%s>.\n", cache, nl_cache_name(cache));
 
@@ -187,27 +187,81 @@ struct nl_cache *nl_cache_alloc(struct nl_cache_ops *ops)
 }
 
 /**
- * Allocate an empty cache based on type name
- * @arg kind		Name of cache type
- * @return A newly allocated and initialized cache.
+ * Allocate new cache and fill it
+ * @arg ops		Cache operations
+ * @arg sock		Netlink socket
+ * @arg result		Result pointer
+ *
+ * Allocate new cache and fill it. Equivalent to calling:
+ * @code
+ * cache = nl_cache_alloc(ops);
+ * nl_cache_refill(sock, cache);
+ * @endcode
+ *
+ * @see nl_cache_alloc
+ *
+ * @return 0 on success or a negative error code.
  */
-struct nl_cache *nl_cache_alloc_name(const char *kind)
+int nl_cache_alloc_and_fill(struct nl_cache_ops *ops, struct nl_sock *sock,
+			    struct nl_cache **result)
 {
-	struct nl_cache_ops *ops;
+	struct nl_cache *cache;
+	int err;
+	
+	if (!(cache = nl_cache_alloc(ops)))
+		return -NLE_NOMEM;
 
-	ops = nl_cache_ops_lookup(kind);
-	if (!ops) {
-		nl_error(ENOENT, "Unable to lookup cache \"%s\"", kind);
-		return NULL;
+	if (sock && (err = nl_cache_refill(sock, cache)) < 0) {
+		nl_cache_free(cache);
+		return err;
 	}
 
-	return nl_cache_alloc(ops);
+	*result = cache;
+	return 0;
 }
 
 /**
- * Allocate a new cache containing a subset of a cache
- * @arg orig		Original cache to be based on
- * @arg filter		Filter defining the subset to be filled into new cache
+ * Allocate new cache based on type name
+ * @arg kind		Name of cache type
+ * @arg result		Result pointer
+ *
+ * Lookup cache ops via nl_cache_ops_lookup() and allocate the cache
+ * by calling nl_cache_alloc(). Stores the allocated cache in the
+ * result pointer provided.
+ *
+ * @see nl_cache_alloc
+ *
+ * @return 0 on success or a negative error code.
+ */
+int nl_cache_alloc_name(const char *kind, struct nl_cache **result)
+{
+	struct nl_cache_ops *ops;
+	struct nl_cache *cache;
+
+	ops = nl_cache_ops_lookup(kind);
+	if (!ops)
+		return -NLE_NOCACHE;
+
+	if (!(cache = nl_cache_alloc(ops)))
+		return -NLE_NOMEM;
+
+	*result = cache;
+	return 0;
+}
+
+/**
+ * Allocate new cache containing a subset of an existing cache
+ * @arg orig		Original cache to base new cache on
+ * @arg filter		Filter defining the subset to be filled into the new cache
+ *
+ * Allocates a new cache matching the type of the cache specified by
+ * \p orig. Iterates over the \p orig cache applying the specified
+ * \p filter and copies all objects that match to the new cache.
+ *
+ * The copied objects are clones but do not contain a reference to each
+ * other. Later modifications to objects in the original cache will
+ * not affect objects in the new cache.
+ *
  * @return A newly allocated cache or NULL.
  */
 struct nl_cache *nl_cache_subset(struct nl_cache *orig,
@@ -234,10 +288,15 @@ struct nl_cache *nl_cache_subset(struct nl_cache *orig,
 }
 
 /**
- * Clear a cache.
- * @arg cache		cache to clear
+ * Remove all objects of a cache.
+ * @arg cache		Cache to clear
  *
- * Removes all elements of a cache.
+ * The objects are unliked/removed from the cache by calling
+ * nl_cache_remove() on each object in the cache. If any of the objects
+ * to not contain any further references to them, those objects will
+ * be freed.
+ *
+ * Unlike with nl_cache_free(), the cache is not freed just emptied.
  */
 void nl_cache_clear(struct nl_cache *cache)
 {
@@ -249,42 +308,23 @@ void nl_cache_clear(struct nl_cache *cache)
 		nl_cache_remove(obj);
 }
 
-static void __nl_cache_free(struct nl_cache *cache)
-{
-	nl_cache_clear(cache);
-
-	NL_DBG(1, "Freeing cache %p <%s>...\n", cache, nl_cache_name(cache));
-	free(cache);
-}
-
-/**
- * Increase reference counter of cache
- * @arg cache		Cache
- */
-void nl_cache_get(struct nl_cache *cache)
-{
-	cache->c_refcnt++;
-}
-
 /**
  * Free a cache.
  * @arg cache		Cache to free.
  *
- * Removes all elements of a cache and frees all memory.
+ * Calls nl_cache_clear() to remove all objects associated with the
+ * cache and frees the cache afterwards.
  *
- * @note Use this function if you are working with allocated caches.
+ * @see nl_cache_clear()
  */
 void nl_cache_free(struct nl_cache *cache)
 {
 	if (!cache)
 		return;
 
-	cache->c_refcnt--;
-	NL_DBG(4, "Returned cache reference %p, %d remaining\n",
-	       cache, cache->c_refcnt);
-
-	if (cache->c_refcnt <= 0)
-		__nl_cache_free(cache);
+	nl_cache_clear(cache);
+	NL_DBG(1, "Freeing cache %p <%s>...\n", cache, nl_cache_name(cache));
+	free(cache);
 }
 
 /** @} */
@@ -308,12 +348,24 @@ static int __cache_add(struct nl_cache *cache, struct nl_object *obj)
 }
 
 /**
- * Add object to a cache.
- * @arg cache		Cache to add object to
+ * Add object to cache.
+ * @arg cache		Cache
  * @arg obj		Object to be added to the cache
  *
- * Adds the given object to the specified cache. The object is cloned
- * if it has been added to another cache already.
+ * Adds the object \p obj to the specified \p cache. In case the object
+ * is already associated with another cache, the object is cloned before
+ * adding it to the cache. In this case, the sole reference to the object
+ * will be the one of the cache. Therefore clearing/freeing the cache
+ * will result in the object being freed again.
+ *
+ * If the object has not been associated with a cache yet, the reference
+ * counter of the object is incremented to account for the additional
+ * reference.
+ *
+ * The type of the object and cache must match, otherwise an error is
+ * returned (-NLE_OBJ_MISMATCH).
+ *
+ * @see nl_cache_move()
  *
  * @return 0 or a negative error code.
  */
@@ -322,12 +374,12 @@ int nl_cache_add(struct nl_cache *cache, struct nl_object *obj)
 	struct nl_object *new;
 
 	if (cache->c_ops->co_obj_ops != obj->ce_ops)
-		return nl_error(EINVAL, "Object mismatches cache type");
+		return -NLE_OBJ_MISMATCH;
 
 	if (!nl_list_empty(&obj->ce_list)) {
 		new = nl_object_clone(obj);
 		if (!new)
-			return nl_errno(ENOMEM);
+			return -NLE_NOMEM;
 	} else {
 		nl_object_get(obj);
 		new = obj;
@@ -341,15 +393,23 @@ int nl_cache_add(struct nl_cache *cache, struct nl_object *obj)
  * @arg cache		Cache to move object to.
  * @arg obj		Object subject to be moved
  *
- * Removes the given object from its associated cache if needed
- * and adds it to the new cache.
+ * Removes the the specified object \p obj from its associated cache
+ * and moves it to another cache.
+ *
+ * If the object is not associated with a cache, the function behaves
+ * just like nl_cache_add().
+ *
+ * The type of the object and cache must match, otherwise an error is
+ * returned (-NLE_OBJ_MISMATCH).
+ *
+ * @see nl_cache_add()
  *
  * @return 0 on success or a negative error code.
  */
 int nl_cache_move(struct nl_cache *cache, struct nl_object *obj)
 {
 	if (cache->c_ops->co_obj_ops != obj->ce_ops)
-		return nl_error(EINVAL, "Object mismatches cache type");
+		return -NLE_OBJ_MISMATCH;
 
 	NL_DBG(3, "Moving object %p to cache %p\n", obj, cache);
 	
@@ -364,12 +424,14 @@ int nl_cache_move(struct nl_cache *cache, struct nl_object *obj)
 }
 
 /**
- * Removes an object from a cache.
- * @arg obj		Object to remove from its cache
+ * Remove object from cache.
+ * @arg obj		Object to remove from cache
  *
- * Removes the object \c obj from the cache it is assigned to, since
- * an object can only be assigned to one cache at a time, the cache
- * must ne be passed along with it.
+ * Removes the object \c obj from the cache it is associated with. The
+ * reference counter of the object will be decremented. If the reference
+ * to the object was the only one remaining, the object will be freed.
+ *
+ * If no cache is associated with the object, this function is a NOP.
  */
 void nl_cache_remove(struct nl_object *obj)
 {
@@ -387,33 +449,6 @@ void nl_cache_remove(struct nl_object *obj)
 	       obj, cache, nl_cache_name(cache));
 }
 
-/**
- * Search for an object in a cache
- * @arg cache		Cache to search in.
- * @arg needle		Object to look for.
- *
- * Iterates over the cache and looks for an object with identical
- * identifiers as the needle.
- *
- * @return Reference to object or NULL if not found.
- * @note The returned object must be returned via nl_object_put().
- */
-struct nl_object *nl_cache_search(struct nl_cache *cache,
-				  struct nl_object *needle)
-{
-	struct nl_object *obj;
-
-	nl_list_for_each_entry(obj, &cache->c_items, ce_list) {
-		if (nl_object_identical(obj, needle)) {
-			nl_object_get(obj);
-			return obj;
-		}
-	}
-
-	return NULL;
-}
-
-
 /** @} */
 
 /**
@@ -422,25 +457,65 @@ struct nl_object *nl_cache_search(struct nl_cache *cache,
  */
 
 /**
- * Request a full dump from the kernel to fill a cache
- * @arg handle		Netlink handle
- * @arg cache		Cache subjected to be filled.
+ * Set synchronization arg1 of cache
+ * @arg cache		Cache
+ * @arg arg		argument
  *
- * Send a dumping request to the kernel causing it to dump all objects
- * related to the specified cache to the netlink socket.
- *
- * Use nl_cache_pickup() to read the objects from the socket and fill them
- * into a cache.
+ * Synchronization arguments are used to specify filters when
+ * requesting dumps from the kernel.
  */
-int nl_cache_request_full_dump(struct nl_handle *handle, struct nl_cache *cache)
+void nl_cache_set_arg1(struct nl_cache *cache, int arg)
+{
+        cache->c_iarg1 = arg;
+}
+
+/**
+ * Set synchronization arg2 of cache
+ * @arg cache		Cache
+ * @arg arg		argument
+ *
+ * Synchronization arguments are used to specify filters when
+ * requesting dumps from the kernel.
+ */
+void nl_cache_set_arg2(struct nl_cache *cache, int arg)
+{
+        cache->c_iarg2 = arg;
+}
+
+/**
+ * Invoke the request-update operation
+ * @arg sk		Netlink socket.
+ * @arg cache		Cache
+ *
+ * This function causes the \e request-update function of the cache
+ * operations to be invoked. This usually causes a dump request to
+ * be sent over the netlink socket which triggers the kernel to dump
+ * all objects of a specific type to be dumped onto the netlink
+ * socket for pickup.
+ *
+ * The behaviour of this function depends on the implemenation of
+ * the \e request_update function of each individual type of cache.
+ *
+ * This function will not have any effects on the cache (unless the
+ * request_update implementation of the cache operations does so).
+ *
+ * Use nl_cache_pickup() to pick-up (read) the objects from the socket
+ * and fill them into the cache.
+ *
+ * @see nl_cache_pickup(), nl_cache_resync()
+ *
+ * @return 0 on success or a negative error code.
+ */
+static int nl_cache_request_full_dump(struct nl_sock *sk,
+				      struct nl_cache *cache)
 {
 	NL_DBG(2, "Requesting dump from kernel for cache %p <%s>...\n",
 	          cache, nl_cache_name(cache));
 
 	if (cache->c_ops->co_request_update == NULL)
-		return nl_error(EOPNOTSUPP, "Operation not supported");
+		return -NLE_OPNOTSUPP;
 
-	return cache->c_ops->co_request_update(cache, handle);
+	return cache->c_ops->co_request_update(cache, sk);
 }
 
 /** @cond SKIP */
@@ -457,8 +532,14 @@ static int update_msg_parser(struct nl_msg *msg, void *arg)
 }
 /** @endcond */
 
-int __cache_pickup(struct nl_handle *handle, struct nl_cache *cache,
-		   struct nl_parser_param *param)
+/**
+ * Pick-up a netlink request-update with your own parser
+ * @arg sk		Netlink socket
+ * @arg cache		Cache
+ * @arg param		Parser parameters
+ */
+static int __cache_pickup(struct nl_sock *sk, struct nl_cache *cache,
+			  struct nl_parser_param *param)
 {
 	int err;
 	struct nl_cb *cb;
@@ -470,17 +551,17 @@ int __cache_pickup(struct nl_handle *handle, struct nl_cache *cache,
 	NL_DBG(1, "Picking up answer for cache %p <%s>...\n",
 		  cache, nl_cache_name(cache));
 
-	cb = nl_cb_clone(handle->h_cb);
+	cb = nl_cb_clone(sk->s_cb);
 	if (cb == NULL)
-		return nl_get_errno();
+		return -NLE_NOMEM;
 
 	nl_cb_set(cb, NL_CB_VALID, NL_CB_CUSTOM, update_msg_parser, &x);
 
-	err = nl_recvmsgs(handle, cb);
+	err = nl_recvmsgs(sk, cb);
 	if (err < 0)
 		NL_DBG(2, "While picking up for %p <%s>, recvmsgs() returned " \
 		       "%d: %s", cache, nl_cache_name(cache),
-		       err, nl_geterror());
+		       err, nl_geterror(err));
 
 	nl_cb_put(cb);
 
@@ -494,7 +575,7 @@ static int pickup_cb(struct nl_object *c, struct nl_parser_param *p)
 
 /**
  * Pickup a netlink dump response and put it into a cache.
- * @arg handle		Netlink handle.
+ * @arg sk		Netlink socket.
  * @arg cache		Cache to put items into.
  *
  * Waits for netlink messages to arrive, parses them and puts them into
@@ -502,18 +583,18 @@ static int pickup_cb(struct nl_object *c, struct nl_parser_param *p)
  *
  * @return 0 on success or a negative error code.
  */
-int nl_cache_pickup(struct nl_handle *handle, struct nl_cache *cache)
+int nl_cache_pickup(struct nl_sock *sk, struct nl_cache *cache)
 {
 	struct nl_parser_param p = {
 		.pp_cb = pickup_cb,
 		.pp_arg = cache,
 	};
 
-	return __cache_pickup(handle, cache, &p);
+	return __cache_pickup(sk, cache, &p);
 }
 
 static int cache_include(struct nl_cache *cache, struct nl_object *obj,
-			 struct nl_msgtype *type, change_func_t cb)
+			 struct nl_msgtype *type, change_func_t cb, void *data)
 {
 	struct nl_object *old;
 
@@ -525,7 +606,7 @@ static int cache_include(struct nl_cache *cache, struct nl_object *obj,
 			nl_cache_remove(old);
 			if (type->mt_act == NL_ACT_DEL) {
 				if (cb)
-					cb(cache, old, NL_ACT_DEL);
+					cb(cache, old, NL_ACT_DEL, data);
 				nl_object_put(old);
 			}
 		}
@@ -533,10 +614,10 @@ static int cache_include(struct nl_cache *cache, struct nl_object *obj,
 		if (type->mt_act == NL_ACT_NEW) {
 			nl_cache_move(cache, obj);
 			if (old == NULL && cb)
-				cb(cache, obj, NL_ACT_NEW);
+				cb(cache, obj, NL_ACT_NEW, data);
 			else if (old) {
 				if (nl_object_diff(old, obj) && cb)
-					cb(cache, obj, NL_ACT_CHANGE);
+					cb(cache, obj, NL_ACT_CHANGE, data);
 
 				nl_object_put(old);
 			}
@@ -551,36 +632,37 @@ static int cache_include(struct nl_cache *cache, struct nl_object *obj,
 }
 
 int nl_cache_include(struct nl_cache *cache, struct nl_object *obj,
-		     change_func_t change_cb)
+		     change_func_t change_cb, void *data)
 {
 	struct nl_cache_ops *ops = cache->c_ops;
 	int i;
 
 	if (ops->co_obj_ops != obj->ce_ops)
-		return nl_error(EINVAL, "Object mismatches cache type");
+		return -NLE_OBJ_MISMATCH;
 
 	for (i = 0; ops->co_msgtypes[i].mt_id >= 0; i++)
 		if (ops->co_msgtypes[i].mt_id == obj->ce_msgtype)
 			return cache_include(cache, obj, &ops->co_msgtypes[i],
-					     change_cb);
+					     change_cb, data);
 
-	return nl_errno(EINVAL);
+	return -NLE_MSGTYPE_NOSUPPORT;
 }
 
 static int resync_cb(struct nl_object *c, struct nl_parser_param *p)
 {
 	struct nl_cache_assoc *ca = p->pp_arg;
 
-	return nl_cache_include(ca->ca_cache, c, ca->ca_change);
+	return nl_cache_include(ca->ca_cache, c, ca->ca_change, ca->ca_change_data);
 }
 
-int nl_cache_resync(struct nl_handle *handle, struct nl_cache *cache,
-		    change_func_t change_cb)
+int nl_cache_resync(struct nl_sock *sk, struct nl_cache *cache,
+		    change_func_t change_cb, void *data)
 {
 	struct nl_object *obj, *next;
 	struct nl_cache_assoc ca = {
 		.ca_cache = cache,
 		.ca_change = change_cb,
+		.ca_change_data = data,
 	};
 	struct nl_parser_param p = {
 		.pp_cb = resync_cb,
@@ -590,20 +672,29 @@ int nl_cache_resync(struct nl_handle *handle, struct nl_cache *cache,
 
 	NL_DBG(1, "Resyncing cache %p <%s>...\n", cache, nl_cache_name(cache));
 
+restart:
 	/* Mark all objects so we can see if some of them are obsolete */
 	nl_cache_mark_all(cache);
 
-	err = nl_cache_request_full_dump(handle, cache);
+	err = nl_cache_request_full_dump(sk, cache);
 	if (err < 0)
 		goto errout;
 
-	err = __cache_pickup(handle, cache, &p);
-	if (err < 0)
+	err = __cache_pickup(sk, cache, &p);
+	if (err == -NLE_DUMP_INTR)
+		goto restart;
+	else if (err < 0)
 		goto errout;
 
-	nl_list_for_each_entry_safe(obj, next, &cache->c_items, ce_list)
-		if (nl_object_is_marked(obj))
+	nl_list_for_each_entry_safe(obj, next, &cache->c_items, ce_list) {
+		if (nl_object_is_marked(obj)) {
+			nl_object_get(obj);
 			nl_cache_remove(obj);
+			if (change_cb)
+				change_cb(cache, obj, NL_ACT_DEL, data);
+			nl_object_put(obj);
+		}
+	}
 
 	NL_DBG(1, "Finished resyncing %p <%s>\n", cache, nl_cache_name(cache));
 
@@ -625,23 +716,19 @@ int nl_cache_parse(struct nl_cache_ops *ops, struct sockaddr_nl *who,
 {
 	int i, err;
 
-	if (!nlmsg_valid_hdr(nlh, ops->co_hdrsize)) {
-		err = nl_error(EINVAL, "netlink message too short to be "
-				       "of kind %s", ops->co_name);
-		goto errout;
-	}
+	if (!nlmsg_valid_hdr(nlh, ops->co_hdrsize))
+		return -NLE_MSG_TOOSHORT;
 
 	for (i = 0; ops->co_msgtypes[i].mt_id >= 0; i++) {
 		if (ops->co_msgtypes[i].mt_id == nlh->nlmsg_type) {
 			err = ops->co_msg_parser(ops, who, nlh, params);
-			if (err != -ENOENT)
+			if (err != -NLE_OPNOTSUPP)
 				goto errout;
 		}
 	}
 
 
-	err = nl_error(EINVAL, "Unsupported netlink message type %d",
-		       nlh->nlmsg_type);
+	err = -NLE_MSGTYPE_NOSUPPORT;
 errout:
 	return err;
 }
@@ -669,7 +756,7 @@ int nl_cache_parse_and_add(struct nl_cache *cache, struct nl_msg *msg)
 
 /**
  * (Re)fill a cache with the contents in the kernel.
- * @arg handle		netlink handle
+ * @arg sk		Netlink socket.
  * @arg cache		cache to update
  *
  * Clears the specified cache and fills it with the current state in
@@ -677,11 +764,12 @@ int nl_cache_parse_and_add(struct nl_cache *cache, struct nl_msg *msg)
  *
  * @return 0 or a negative error code.
  */
-int nl_cache_refill(struct nl_handle *handle, struct nl_cache *cache)
+int nl_cache_refill(struct nl_sock *sk, struct nl_cache *cache)
 {
 	int err;
 
-	err = nl_cache_request_full_dump(handle, cache);
+restart:
+	err = nl_cache_request_full_dump(sk, cache);
 	if (err < 0)
 		return err;
 
@@ -689,7 +777,13 @@ int nl_cache_refill(struct nl_handle *handle, struct nl_cache *cache)
 	       cache, nl_cache_name(cache));
 	nl_cache_clear(cache);
 
-	return nl_cache_pickup(handle, cache);
+	err = nl_cache_pickup(sk, cache);
+	if (err == -NLE_DUMP_INTR) {
+		fprintf(stderr, "dump interrupted, restarting!\n");
+		goto restart;
+	}
+	
+	return err;
 }
 
 /** @} */
@@ -700,8 +794,41 @@ int nl_cache_refill(struct nl_handle *handle, struct nl_cache *cache)
  */
 
 /**
- * Mark all objects in a cache
- * @arg cache		Cache to mark all objects in
+ * Search object in cache
+ * @arg cache		Cache
+ * @arg needle		Object to look for.
+ *
+ * Searches the cache for an object which matches the object \p needle.
+ * The function nl_object_identical() is used to determine if the
+ * objects match. If a matching object is found, the reference counter
+ * is incremented and the object is returned.
+ * 
+ * Therefore, if an object is returned, the reference to the object
+ * must be returned by calling nl_object_put() after usage.
+ *
+ * @return Reference to object or NULL if not found.
+ */
+struct nl_object *nl_cache_search(struct nl_cache *cache,
+				  struct nl_object *needle)
+{
+	struct nl_object *obj;
+
+	nl_list_for_each_entry(obj, &cache->c_items, ce_list) {
+		if (nl_object_identical(obj, needle)) {
+			nl_object_get(obj);
+			return obj;
+		}
+	}
+
+	return NULL;
+}
+
+/**
+ * Mark all objects of a cache
+ * @arg cache		Cache
+ *
+ * Marks all objects of a cache by calling nl_object_mark() on each
+ * object associated with the cache.
  */
 void nl_cache_mark_all(struct nl_cache *cache)
 {
@@ -746,7 +873,7 @@ void nl_cache_dump_filter(struct nl_cache *cache,
 			  struct nl_dump_params *params,
 			  struct nl_object *filter)
 {
-	int type = params ? params->dp_type : NL_DUMP_FULL;
+	int type = params ? params->dp_type : NL_DUMP_DETAILS;
 	struct nl_object_ops *ops;
 	struct nl_object *obj;
 
@@ -814,10 +941,22 @@ void nl_cache_foreach_filter(struct nl_cache *cache, struct nl_object *filter,
 		BUG();
 
 	nl_list_for_each_entry_safe(obj, tmp, &cache->c_items, ce_list) {
-		if (filter && !nl_object_match_filter(obj, filter))
-			continue;
+		if (filter) {
+			int diff = nl_object_match_filter(obj, filter);
+
+			NL_DBG(3, "%p<->%p object difference: %x\n",
+				obj, filter, diff);
+
+			if (!diff)
+				continue;
+		}
+
+		/* Caller may hold obj for a long time */
+		nl_object_get(obj);
 
 		cb(obj, arg);
+
+		nl_object_put(obj);
 	}
 }
 
