@@ -125,12 +125,9 @@ static int route_clone(struct nl_object *_dst, struct nl_object *_src)
 static void route_dump_line(struct nl_object *a, struct nl_dump_params *p)
 {
 	struct rtnl_route *r = (struct rtnl_route *) a;
-	struct nl_cache *link_cache;
 	int cache = 0, flags;
 	char buf[64];
 
-	link_cache = nl_cache_mngt_require("route/link");
-
 	if (r->rt_flags & RTM_F_CLONED)
 		cache = 1;
 
@@ -284,73 +281,8 @@ static void route_dump_stats(struct nl_object *obj, struct nl_dump_params *p)
 		nl_dump_line(p, "    used %u refcnt %u last-use %us "
 				"expires %us\n",
 			     ci->rtci_used, ci->rtci_clntref,
-			     ci->rtci_last_use / nl_get_hz(),
-			     ci->rtci_expires / nl_get_hz());
-	}
-}
-
-static void route_dump_env(struct nl_object *obj, struct nl_dump_params *p)
-{
-	struct rtnl_route *route = (struct rtnl_route *) obj;
-	struct nl_cache *link_cache;
-	char buf[128];
-
-	link_cache = nl_cache_mngt_require("route/link");
-
-	nl_dump_line(p, "ROUTE_FAMILY=%s\n",
-		     nl_af2str(route->rt_family, buf, sizeof(buf)));
-
-	if (route->ce_mask & ROUTE_ATTR_DST)
-		nl_dump_line(p, "ROUTE_DST=%s\n",
-			     nl_addr2str(route->rt_dst, buf, sizeof(buf)));
-
-	if (route->ce_mask & ROUTE_ATTR_SRC)
-		nl_dump_line(p, "ROUTE_SRC=%s\n",
-			     nl_addr2str(route->rt_src, buf, sizeof(buf)));
-
-	if (route->ce_mask & ROUTE_ATTR_PREF_SRC)
-		nl_dump_line(p, "ROUTE_PREFSRC=%s\n",
-			     nl_addr2str(route->rt_pref_src, buf, sizeof(buf)));
-
-	if (route->ce_mask & ROUTE_ATTR_IIF) {
-		if (link_cache) {
-			nl_dump_line(p, "ROUTE_IIF=%s",
-				rtnl_link_i2name(link_cache, route->rt_iif,
-						 buf, sizeof(buf)));
-		} else
-			nl_dump_line(p, "ROUTE_IIF=%d", route->rt_iif);
-	}
-
-	if (route->ce_mask & ROUTE_ATTR_TOS)
-		nl_dump_line(p, "ROUTE_TOS=%u\n", route->rt_tos);
-
-	if (route->ce_mask & ROUTE_ATTR_TABLE)
-		nl_dump_line(p, "ROUTE_TABLE=%u\n",
-			     route->rt_table);
-
-	if (route->ce_mask & ROUTE_ATTR_SCOPE)
-		nl_dump_line(p, "ROUTE_SCOPE=%s\n",
-			     rtnl_scope2str(route->rt_scope, buf, sizeof(buf)));
-
-	if (route->ce_mask & ROUTE_ATTR_PRIO)
-		nl_dump_line(p, "ROUTE_PRIORITY=%u\n",
-			     route->rt_prio);
-
-	if (route->ce_mask & ROUTE_ATTR_TYPE)
-		nl_dump_line(p, "ROUTE_TYPE=%s\n",
-			     nl_rtntype2str(route->rt_type, buf, sizeof(buf)));
-
-	if (route->ce_mask & ROUTE_ATTR_MULTIPATH) {
-		struct rtnl_nexthop *nh;
-		int index = 1;
-
-		if (route->rt_nr_nh > 0)
-			nl_dump_line(p, "ROUTE_NR_NH=%u\n", route->rt_nr_nh);
-
-		nl_list_for_each_entry(nh, &route->rt_nexthops, rtnh_list) {
-			p->dp_ivar = index++;
-			rtnl_route_nh_dump(nh, p);
-		}
+			     ci->rtci_last_use / nl_get_user_hz(),
+			     ci->rtci_expires / nl_get_user_hz());
 	}
 }
 
@@ -411,9 +343,10 @@ static int route_compare(struct nl_object *_a, struct nl_object *_b,
 			found = 0;
 			nl_list_for_each_entry(nh_b, &b->rt_nexthops,
 					       rtnh_list) {
-				if (!rtnl_route_nh_compare(nh_a, nh_b, ~0, 0))
+				if (!rtnl_route_nh_compare(nh_a, nh_b, ~0, 0)) {
 					found = 1;
 					break;
+				}
 			}
 			if (!found)
 				goto nh_mismatch;
@@ -425,9 +358,10 @@ static int route_compare(struct nl_object *_a, struct nl_object *_b,
 			found = 0;
 			nl_list_for_each_entry(nh_a, &a->rt_nexthops,
 					       rtnh_list) {
-				if (!rtnl_route_nh_compare(nh_a, nh_b, ~0, 0))
+				if (!rtnl_route_nh_compare(nh_a, nh_b, ~0, 0)) {
 					found = 1;
 					break;
+				}
 			}
 			if (!found)
 				goto nh_mismatch;
@@ -455,7 +389,7 @@ nh_mismatch:
 #undef ROUTE_DIFF
 }
 
-static struct trans_tbl route_attrs[] = {
+static const struct trans_tbl route_attrs[] = {
 	__ADD(ROUTE_ATTR_FAMILY, family)
 	__ADD(ROUTE_ATTR_TOS, tos)
 	__ADD(ROUTE_ATTR_TABLE, table)
@@ -752,18 +686,26 @@ void rtnl_route_add_nexthop(struct rtnl_route *route, struct rtnl_nexthop *nh)
 
 void rtnl_route_remove_nexthop(struct rtnl_route *route, struct rtnl_nexthop *nh)
 {
-	route->rt_nr_nh--;
-	nl_list_del(&nh->rtnh_list);
+	if (route->ce_mask & ROUTE_ATTR_MULTIPATH) {
+		route->rt_nr_nh--;
+		nl_list_del(&nh->rtnh_list);
+	}
 }
 
 struct nl_list_head *rtnl_route_get_nexthops(struct rtnl_route *route)
 {
-	return &route->rt_nexthops;
+	if (route->ce_mask & ROUTE_ATTR_MULTIPATH)
+		return &route->rt_nexthops;
+
+	return NULL;
 }
 
 int rtnl_route_get_nnexthops(struct rtnl_route *route)
 {
-	return route->rt_nr_nh;
+	if (route->ce_mask & ROUTE_ATTR_MULTIPATH)
+		return route->rt_nr_nh;
+
+	return 0;
 }
 
 void rtnl_route_foreach_nexthop(struct rtnl_route *r,
@@ -1136,7 +1078,17 @@ int rtnl_route_build_msg(struct nl_msg *msg, struct rtnl_route *route)
 		nla_nest_end(msg, metrics);
 	}
 
-	if (rtnl_route_get_nnexthops(route) > 0) {
+	if (rtnl_route_get_nnexthops(route) == 1) {
+		struct rtnl_nexthop *nh;
+
+		nh = rtnl_route_nexthop_n(route, 0);
+		if (nh->rtnh_gateway)
+			NLA_PUT_ADDR(msg, RTA_GATEWAY, nh->rtnh_gateway);
+		if (nh->rtnh_ifindex)
+			NLA_PUT_U32(msg, RTA_OIF, nh->rtnh_ifindex);
+		if (nh->rtnh_realms)
+			NLA_PUT_U32(msg, RTA_FLOW, nh->rtnh_realms);
+	} else if (rtnl_route_get_nnexthops(route) > 1) {
 		struct nlattr *multipath;
 		struct rtnl_nexthop *nh;
 
@@ -1185,7 +1137,6 @@ struct nl_object_ops route_obj_ops = {
 	    [NL_DUMP_LINE]	= route_dump_line,
 	    [NL_DUMP_DETAILS]	= route_dump_details,
 	    [NL_DUMP_STATS]	= route_dump_stats,
-	    [NL_DUMP_ENV]	= route_dump_env,
 	},
 	.oo_compare		= route_compare,
 	.oo_attrs2str		= route_attrs2str,
